"""Usage of `setup.py` is deprecated, and is supplied only for legacy installation.
"""
import sys
import os
import os.path as osp
import importlib
import logging
import argparse
import subprocess
import tempfile
from argparse import RawTextHelpFormatter
logger = logging.getLogger(__name__)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def egg_info( args ):

  logger.warning(
    "running legacy 'setup.py egg_info'" )

  dir = osp.join( args.egg_base, EGG_INFO_NAME )

  if not osp.exists( dir ):
    os.mkdir( dir )

  with open( osp.join( dir, 'PKG-INFO' ), 'wb' ) as fp:
    fp.write( PKG_INFO )

  with open( osp.join( dir, 'setup_requires.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'requires.txt' ), 'wb' ) as fp:
    fp.write( REQUIRES )

  with open( osp.join( dir, 'SOURCES.txt' ), 'wb' ) as fp:
    fp.write( SOURCES )

  with open( osp.join( dir, 'top_level.txt' ), 'wb' ) as fp:
    fp.write( TOP_LEVEL )

  with open( osp.join( dir, 'entry_points.txt' ), 'wb' ) as fp:
    fp.write( ENTRY_POINTS )

  with open( osp.join( dir, 'dependency_links.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'not-zip-safe' ), 'wb' ) as fp:
    fp.write( b'' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def bdist_wheel( args ):

  logger.warning(
    "running legacy 'setup.py bdist_wheel'" )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  backend.build_wheel(
    wheel_directory = args.dist_dir or args.bdist_dir or '.' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def install( args ):

  logger.warning(
    "running legacy 'setup.py install'" )

  reqs = [ f"{{r}}" for r in build_requires ]

  subprocess.check_call([
    sys.executable,
    '-m',
    'pip',
    'install',
    *reqs ] )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  with tempfile.TemporaryDirectory() as tmpdir:
    wheel_name = backend.build_wheel(
      wheel_directory = tmpdir )

    subprocess.check_call([
      sys.executable,
      '-m',
      'pip',
      'install',
      osp.join(tmpdir, wheel_name) ])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dummy( args ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def main():

  logging.basicConfig(
    level = logging.INFO,
    format = "{{name}}:{{levelname}}: {{message}}",
    style = "{{" )


  logger.warning(
    "'setup.py' is deprecated, limited support for legacy installs. Upgrade pip." )

  parser = argparse.ArgumentParser(
    description = __doc__,
    formatter_class = RawTextHelpFormatter )

  subparsers = parser.add_subparsers()

  #.............................................................................
  egg_info_parser = subparsers.add_parser( 'egg_info' )

  egg_info_parser.set_defaults( func = egg_info )

  egg_info_parser.add_argument( "-e", "--egg-base",
    type = str,
    default = '.' )

  #.............................................................................
  bdist_wheel_parser = subparsers.add_parser( 'bdist_wheel' )

  bdist_wheel_parser.set_defaults( func = bdist_wheel )

  bdist_wheel_parser.add_argument( "-b", "--bdist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "-d", "--dist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "--python-tag",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--plat-name",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--py-limited-api",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--build-number",
    type = str,
    default = None )

  #.............................................................................
  install_parser = subparsers.add_parser( 'install' )

  install_parser.set_defaults( func = install )

  install_parser.add_argument( "--record",
    type = str,
    default = None )

  install_parser.add_argument( "--install-headers",
    type = str,
    default = None )

  install_parser.add_argument( "--compile",
    action='store_true' )

  install_parser.add_argument( "--single-version-externally-managed",
    action='store_true' )

  #.............................................................................
  clean_parser = subparsers.add_parser( 'clean' )

  clean_parser.set_defaults( func = dummy )

  clean_parser.add_argument( "-a", "--all",
    action='store_true' )

  args = parser.parse_args( )

  args.func( args )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# NOTE: these are templated literal values substituded by the backend when
# building the source distribution

build_backend = {build_backend}
backend_path = {backend_path}
build_requires = {build_requires}

EGG_INFO_NAME = {egg_info_name}

PKG_INFO = {pkg_info}

REQUIRES = {requires}

SOURCES = {sources}

TOP_LEVEL = {top_level}

ENTRY_POINTS = {entry_points}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

if __name__ == "__main__":
  exit( main() )



import os
import os.path as osp
import tempfile

from pytest import (
  raises )

from partis.pyproj import (
  ValidationError,
  PkgInfo,
  PkgInfoAuthor,
  PkgInfoURL,
  PkgInfoReq )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_base():
  assert PkgInfoAuthor('asd') == PkgInfoAuthor('asd')
  assert PkgInfoAuthor('asd') != PkgInfoAuthor('xyz')

  assert PkgInfoURL('asd', 'http://asd.com') == PkgInfoURL('asd', 'http://asd.com')
  assert PkgInfoURL('asd', 'http://asd.com') != PkgInfoURL('xyz', 'http://xyz.com')

  assert PkgInfoReq('asd') == PkgInfoReq('asd')
  assert PkgInfoReq('asd') != PkgInfoReq('xyz')

  PkgInfoReq(
    "PySide2 >= 5.14, < 5.15; python_version < '3.8'",
    extra = 'gui')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_default():
  pkginfo = PkgInfo(
    project = dict(
      name = 'test_pkg',
      version = '1.2.3' ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_full():
  with tempfile.TemporaryDirectory() as tmpdir:

    dynamic = ['dependencies']

    authors = [
      {'name': 'asd'},
      {'email': 'asd@asd.com'},
      {'name': 'asd', 'email': 'asd@asd.com'}]

    keywords = ['axat']

    classifiers = ['asd :: asd']

    urls = {
      'home' : 'http://home.com' }

    project = {
      'name' : 'test_pkg',
      'version' : '1.2.3',
      'description' : "asd",
      # 'readme' : '',
      'authors' : authors,
      'maintainers' : authors,
      # 'license' : None,
      # 'dynamic' : dynamic,
      'requires-python' : ">= 3.6.2",
      'dependencies' : ["numpy"],
      'optional-dependencies' : {
        'test' : 'pytest' },
      'keywords' : keywords,
      'classifiers' : classifiers,
      'urls' : urls,
      'scripts' : {
        'xyz' : 'abc.xyz:func' },
      'gui-scripts': {
        'xyz' : 'abc.xyz:func' },
      'entry-points' : {
        'plugin' : {
          'xyz' : 'abc.xyz:func' } } }

    pkginfo = PkgInfo(
      root = tmpdir,
      project = project )

    pkginfo.add_dependencies(['scipy'])

    #...........................................................................
    readme = "Test Package"

    readme_file_txt = 'readme'
    readme_file_md = 'readme.md'
    readme_file_rst = 'readme.rst'

    with open(osp.join(tmpdir, readme_file_txt), 'w') as fp:
      fp.write(readme)

    with open(osp.join(tmpdir, readme_file_md), 'w') as fp:
      fp.write(readme)

    with open(osp.join(tmpdir, readme_file_rst), 'w') as fp:
      fp.write(readme)

    #...........................................................................
    invalid_entry_points = [
      { 'scripts' : {} },
      { 'console_scripts' : {} },
      { 'gui-scripts' : {} },
      { 'gui_scripts' : {} },
      { 'nested' : {
        'toomuch' : {
          'xyz' : 'abc.xyz:func' } } } ]

    for entry_points in invalid_entry_points:
      with raises( ValidationError ):
        pkginfo = PkgInfo(
          project = {
            **project,
            'entry-points' : entry_points })

    #...........................................................................

    valid_readme = [
      { 'text' : readme },
      { 'file': readme_file_md },
      { 'file': readme_file_rst },
      { 'file': readme_file_txt } ]

    invalid_readme = [
      'junk',
      {},
      { 'text' : readme, 'file': readme_file_md },
      { 'file': 'junk' },
      { 'junk' : 'junk' } ]



    for readme in valid_readme:

      pkginfo = PkgInfo(
        root = tmpdir,
        project = {
          **project,
          'readme' : readme })

      pkginfo.encode_pkg_info()
      pkginfo.encode_entry_points()

    with raises( ValidationError ):
      pkginfo = PkgInfo(
        project = {
          **project,
          'readme' : { 'file': readme_file_md } })

    with raises( ValidationError ):
      pkginfo = PkgInfo(
        project = {
          **project,
          'readme' : readme_file_md })

    for readme in invalid_readme:
      print(readme)

      with raises( ValidationError ):
        pkginfo = PkgInfo(
          root = tmpdir,
          project = {
            **project,
            'readme' : readme })


    #...........................................................................

    valid_license = [
      { 'text' : readme, 'file': readme_file_md },
      { 'text' : readme },
      { 'file': readme_file_md } ]

    invalid_license = [
      'junk',
      {},
      { 'file': 'junk' },
      { 'junk' : 'junk' } ]


    for license in valid_license:

      pkginfo = PkgInfo(
        root = tmpdir,
        project = {
          **project,
          'license' : license })

      pkginfo.encode_pkg_info()
      pkginfo.encode_entry_points()

    with raises( ValidationError ):
      pkginfo = PkgInfo(
        project = {
          **project,
          'license' : { 'file': readme_file_md } })

    for license in invalid_license:
      print(license)

      with raises( ValidationError ):
        pkginfo = PkgInfo(
          root = tmpdir,
          project = {
            **project,
            'license' : license })



import warnings
from copy import copy

from pytest import (
  warns,
  raises )

from partis.pyproj.validate import (
  ValidationError,
  ValidDefinitionError,
  validating,
  validate,
  Optional,
  optional,
  Required,
  required,
  fmt_validator,
  Validator,
  Restricted,
  valid,
  union,
  restrict,
  valid_type,
  valid_keys,
  valid_dict,
  valid_list,
  mapget,
  as_list )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_special():
  assert optional == Optional()
  print(str(optional))
  print(repr(optional))
  print(hash(optional))

  assert required == Required()
  print(hash(required))

  assert optional != required
  assert optional != None
  assert required != None

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_validating():

  def f():
    assert False

  def g():
    raise ValidationError("qwe")

  with raises( ValidationError ):
    try:
      with validating( key = 1, root = {1:'a'}, file = 'asd'):
        with validating( key = 'xyz' ):
          with validating( key = 2 ):
            f()

    except ValidationError as e:
      assert e.msg == "Error while validating"
      assert e.doc_path == [1, 'xyz', 2]
      assert e.doc_root == {1:'a'}
      assert e.doc_file == 'asd'
      print(str(e))
      print(repr(e))
      raise

  with raises( ValidationError ):
    try:
      with validating( key = 'xyz', root = {1:'a'}, file = 'asd'):
        with validating( key = 1 ):
          g()

    except ValidationError as e:
      assert e.msg == "qwe"
      assert e.doc_path == ['xyz', 1]
      assert e.doc_root == {1:'a'}
      assert e.doc_file == 'asd'
      print(str(e))
      raise


  with raises( ValidationError ):
    try:
      with validating():
        g()

    except ValidationError as e:
      assert e.msg == "qwe"
      assert e.doc_path == list()
      assert e.doc_root is None
      assert e.doc_file is None
      print(str(e))
      raise

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_validate():

  assert validate(None, 1.0, [float]) == 1.0
  assert validate(None, 1.0, float) == 1.0

  with raises( ValidationError ):
    validate(None, required, [int])

  with raises( ValidationError ):
    validate('asd', required, [int])

  def f(x):
    raise ValidationError("asd")

  with raises( ValidationError ):
    validate('asd', required, [f])

  assert validate('12.34', required, [[int, float]]) == 12.34
  assert validate(12.34, required, [[int, float]]) == 12
  assert validate('asdasd', required, [[]]) == 'asdasd'

  with raises( ValidationError ):
    validate('12.34', required, [[int, f]])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_validator():

  m = Validator()
  print(str(m))
  print(repr(m))
  assert m._default == required
  assert m._validators == []

  m = Validator(default = None)
  assert m._default == optional
  assert m._validators == []

  m = Validator(None)
  assert m._default == optional
  assert m._validators == []

  a = Validator(1)
  print(str(a))
  assert a._default == 1
  assert a._validators == [int]

  b = Validator(int)
  print(str(b))
  assert b._default == 0
  assert b._validators == [int]

  c = Validator(int, default = 2)
  print(str(c))
  assert c._default == 2
  assert c._validators == [int]

  class Test:
    def __init__(self, x):
      self._x = x

    def __eq__(self, other):
      if isinstance(other, Test):

        return self._x == other._x

      return self._x == other

  d = Validator(Test, default = 1)
  print(str(d))
  assert d._default == Test(1)
  assert d._validators == [Test]

  with raises( ValidDefinitionError ):
    Validator(Test)

  def f(x):
    return x

  print(str(Validator(1, int, a, f, lambda x: x)))
  print(str(Validator('')))

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_restricted():
  a = Restricted(1, 5, 10)
  assert a._default == 1
  assert a._validators == [int]
  assert a._options == {1, 5, 10}

  assert a(1) == 1
  assert a(5) == 5
  assert a(10) == 10

  with raises( ValidationError ):
    a(2)

  with raises( ValidDefinitionError ):
    Restricted()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valid_type():
  valid_type('xyz', types = [str])

  with raises( ValidationError ):
    valid_type('xyz', types = [int])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valid_keys():
  with raises( ValidationError ):
    valid_keys( 'xyz', allow_keys = ['x', 'y', 'z'] )

  with raises( ValidationError ):
    valid_keys( list(), allow_keys = ['x', 'y', 'z'] )

  valid_keys( dict(), allow_keys = ['x', 'y', 'z'] )

  obj = {
    'x': 1,
    'y': 2,
    'z': 3 }

  valid_keys( obj, allow_keys = ['w', 'x', 'y', 'z'] )

  with raises( ValidationError ):
    valid_keys( obj, allow_keys = ['x', 'y'] )

  with raises( ValidationError ):
    valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      require_keys = ['w'] )

  with raises( ValidationError ):
    valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      mutex_keys = [('x', 'y')] )

  with raises( ValidationError ):
    valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      deprecate_keys = [('x', required)] )

  obj2 = valid_keys( obj,
    allow_keys = ['x', 'y', 'z'],
    deprecate_keys = [('x', None)] )

  assert 'x' not in obj2

  with warns(DeprecationWarning):
    obj2 = valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      deprecate_keys = [('x', 'w')] )

  assert 'x' not in obj2
  assert obj2['w'] == 1

  with warns(DeprecationWarning):
    obj2 = valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      deprecate_keys = [('x', 'y')] )

  assert 'x' not in obj2
  assert obj2['y'] == 2

  with warnings.catch_warnings():
    warnings.simplefilter("error")

    # should not warn because obj does not have key 'w'
    valid_keys( obj,
      allow_keys = ['x', 'y', 'z'],
      deprecate_keys = [('w', None)] )

  with raises( ValidationError ):
    valid_keys( obj,
      allow_keys = ['y', 'z'],
      wedge_keys = [('w', 'x')] )

  valid_keys( obj,
    allow_keys = ['z'],
    wedge_keys = [('x', 'y')] )

  _obj = {
    ' x': 1,
    'y ': 2,
    ' z   ': 3 }

  obj2 = valid_keys( _obj,
    key_valid = lambda k: k.strip() )

  assert obj2 == obj
  assert obj2 != _obj

  _obj = {
    'x': 1,
    'y': 2.2,
    'z': 3 }

  obj2 = valid_keys( _obj,
    value_valid = lambda v: int(v) )

  assert obj2 is not _obj
  assert obj2 == obj
  assert obj2 != _obj

  _obj = {
    ' x': 1,
    'y': 2.2,
    ' z   ': 3 }

  obj2 = valid_keys( _obj,
    item_valid = lambda kv: (kv[0].strip(), int(kv[1]) ) )

  assert obj2 is not _obj
  assert obj2 == obj
  assert obj2 != _obj

  _obj = {
    'x': 1 }

  obj2 = valid_keys( _obj,
    allow_keys = list(),
    proxy_keys = [('y', 'x'), ('z', 'x')] )


  assert obj2['x'] == 1
  assert obj2['y'] == _obj['x']
  assert obj2['z'] == _obj['x']

  obj2 = valid_keys( {},
    default = obj )

  assert obj2 is not _obj
  assert obj2 == obj


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valid_dict():
  class test(valid_dict):
    _deprecate_keys = [('x', None), ('y', required)]
    _wedge_keys = [('a','b')]
    _mutex_keys = [('c', 'd')]

  a = test()

  with warns(DeprecationWarning):
    a.setdefault('x', 4)

  with raises(ValidationError):
    a['y'] = 6

  b = test({'a':1, 'b':2, 'c':3})

  c = copy(b)
  print(str(b))
  print(repr(b))
  print(str(c))

  for k in b:
    print(k)

  for k in b.keys():
    print(k)

  for v in b.values():
    print(v)

  assert c.pop('c') == 3
  assert 'c' not in c

  class test(valid_dict):
    _proxy_key = 'x'

  a = test()
  assert a == {}

  b = test(None)
  assert b == {}

  c = test(1)
  assert 'x' in c
  assert c['x'] == 1



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valid_list():
  class test(valid_list):
    pass

  a = test()
  a.extend([1,2,3])
  a[1] = 4
  print(str(a))
  print(repr(a))

  a.clear()

  class test(valid_list):
    _min_len = 1

  a = test(['a', 'b'])
  assert a.pop(0) == 'a'

  with raises(ValidationError):
    a.pop()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_mapget():
  obj = {
    'x': 1,
    'y': {
      'a' : 2,
      'b' : [4,5,6] },
    'z': [7,8,9] }

  default = 1234

  valid_paths = [
    ('w', default),
    ('x', 1),
    ('y.a', 2),
    ('y.b', [4,5,6]),
    ('y.c', default),
    ('z', [7,8,9]) ]

  invalid_paths = [
    'x.y',
    'y.a.c',
    'y.b.0.z',
    'z.0' ]

  for p, val in valid_paths:
    assert val == mapget( obj, p, default = default )

  for p in invalid_paths:
    print(p)
    with raises( ValidationError ):
      mapget( obj, p, default = default )

  with raises( ValidationError ):
    mapget( 'asd', 'x', default = default )



import os
import os.path as osp
import glob
import logging
import re
from fnmatch import (
  translate )
import posixpath

from ..validate import (
  ValidationError,
  FileOutsideRootError,
  validating )

from ..norms import (
  norm_path )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def norm_join_base(base, dirname, names):
  """Creates paths relative to a 'base' path for a list of names in a 'dirname'
  """

  rpath = osp.normcase( osp.relpath( dirname, start = base ) )

  if osp is posixpath:
    return [
      (name, osp.normpath( osp.join(rpath, name) ))
      for name in names ]
  else:
    return [
      (name, osp.normpath( osp.join(rpath, osp.normcase(name)) ))
      for name in names ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def compile_patterns(base, patterns):
  """Creates compiled glob patterns
  """
  patterns = [ osp.normcase(p) for p in patterns ]

  # create mask for patterns containing path separator
  # these are clearly intended to be matched against the full path name of the file.
  base_mask = [ p.find(osp.sep) >= 0 for p in patterns ]

  return base_mask, [
    re.compile( translate(p) ).match
    for p in patterns ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def match_names(name_paths, base_mask, pattern_matches):
  """Matches filenames base on either the basename, or path relative to base path,
  depending on if the compiled pattern was detected to be for a single name or
  a path
  """

  filtered_names = []

  for mask, match in zip(base_mask, pattern_matches):
    if mask:
      filtered_names.extend([ name for name, path in name_paths if match(path) ])
    else:
      filtered_names.extend([ name for name, path in name_paths if match(name) ])

  return filtered_names

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def root_ignore_patterns(
  *patterns,
  base,
  root ):
  """Create ignore function that checks for files that don't have common path with root

  Parameters
  ----------
  *patterns : str
    Patterns used to match ignore files using :mod:`fnmatch `
  base : str
    Base path used for path-like ignore patterns (those containing a path separator)
  root : str
    Base path to check real paths against. Any files found that do not have a common
    path with root, and which would not otherwise be ignored,
    will raise an ``FileOutsideRootError``.

  Raises
  ------
  FileOutsideRootError

  See Also
  --------
  * :func:`shutil.ignore_patterns` : This factory function creates a function
    that can be used as a callable for copytree()'s ignore argument, ignoring
    files and directories that match one of the glob-style patterns provided.

  """

  base = osp.normpath(os.fspath(base))
  root = osp.normpath(os.fspath(root))

  base_mask, pattern_matches = compile_patterns(base, patterns)

  def _ignore_patterns(path, names):
    path = osp.normpath(os.fspath(path))

    name_paths = norm_join_base( base, path, names )
    ignored_names = match_names(name_paths, base_mask, pattern_matches)

    # debugging prints (no logging)
    # print('base: ', base)
    # print('patterns: ', list(zip(base_mask,patterns)))
    # print('name_paths: ', name_paths)
    # print('ignored_names: ', ignored_names)

    for name in names:
      if name not in ignored_names:
        abs_path = osp.realpath( osp.join(
          path,
          name ) )

        prefix = osp.commonpath([root, abs_path])

        if prefix != root:
          raise FileOutsideRootError(
            f"Must have common path with root:\n  file = \"{abs_path}\"\n  root = \"{root}\"")

    return set(ignored_names)

  return _ignore_patterns

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def combine_ignore_patterns(*ignores):

  def _ignore_patterns(path, names):
    ignored_names = set()

    for ignore in ignores:
      ignored_names |= ignore(path, names)

    return ignored_names

  return _ignore_patterns

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dist_iter(*,
  include,
  ignore,
  root ):

  ignore_patterns = root_ignore_patterns(
    *ignore,
    base = '.',
    root = root )

  for i, incl in enumerate(include):
    src = incl.src
    dst = incl.dst
    _ignore = incl.ignore

    _ignore_patterns = combine_ignore_patterns(
      ignore_patterns,
      root_ignore_patterns(
        *_ignore,
        base = src,
        root = root ) )

    if incl.glob:

      cwd = os.getcwd()
      try:
        os.chdir(src)
        matches = glob.glob(incl.glob, recursive = True)
      finally:
        os.chdir(cwd)

      for match in matches:
        _src = osp.join(src, match)
        # re-base the dst path, path relative to src == path relative to dst
        _dst = osp.join(dst, match)

        yield ( i, _src, _dst, _ignore_patterns, False )

    else:

      yield ( i, src, dst, _ignore_patterns, True )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dist_copy(*,
  base_path,
  include,
  ignore,
  dist,
  root = None,
  logger = None ):

  if len(include) == 0:
    return

  logger = logger or logging.getLogger( __name__ )

  with validating(key = 'copy'):

    for i, src, dst, ignore_patterns, individual in dist_iter(
      include = include,
      ignore = ignore,
      root = root ):

      with validating(key = i):

        src = osp.normpath( src )
        dst = '/'.join( [base_path, norm_path(dst)] )

        ignored = ignore_patterns(
          osp.dirname(src),
          [ osp.basename(src) ])

        if ignored and not individual:
          logger.debug( f'ignoring: {src}' )
          continue

        logger.debug(f"dist copy: {src} -> {dst}")

        if osp.isdir( src ):
          dist.copytree(
            src = src,
            dst = dst,
            ignore = ignore_patterns )

        else:
          dist.copyfile(
            src = src,
            dst = dst )



import os
import os.path as osp
import sys
import shutil
import logging
import tempfile
from copy import copy, deepcopy
from collections.abc import (
  Mapping,
  Sequence )
import subprocess
import multiprocessing
import glob
import warnings
import tomli

try:
  from importlib.metadata import metadata

except ImportError:
  from importlib_metadata import metadata

from .pkginfo import (
  PkgInfoReq,
  PkgInfo )

from .validate import (
  ValidationWarning,
  ValidationError,
  FileOutsideRootError,
  valid_dict,
  validating,
  valid,
  restrict,
  mapget )

from .norms import (
  scalar_list,
  norm_bool,
  norm_path_to_os,
  norm_path )

from .pep import (
  purelib_compat_tags,
  platlib_compat_tags )

from .load_module import (
  EntryPointError,
  load_module,
  load_entrypoint )

from .legacy import legacy_setup_content

from .pptoml import (
  pptoml )

from .meson import (
  MesonBuild )

from .dist_file import (
  dist_copy )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class PyProjBase:
  """Minimal build system for a Python project

  Extends beyond :pep:`517` and :pep:`621`


  Parameters
  ----------
  root : str
    Path to the root project directory containing 'pyproject.toml'.
  logger : logging.Logger
    Parent logger to use when processing project.

  """
  #-----------------------------------------------------------------------------
  def __init__( self, *,
    root,
    config_settings = None,
    logger = None ):

    self.logger = logger or logging.getLogger( __name__ )

    self.root = osp.realpath(root)

    self.pptoml_file = osp.join( self.root, 'pyproject.toml' )

    with open( self.pptoml_file, 'rb' ) as fp:
      src = fp.read()
      src = src.decode( 'utf-8', errors = 'replace' )
      self._pptoml = tomli.loads( src )

    with validating(root = self._pptoml, file = self.pptoml_file):
      self._pptoml = pptoml(self._pptoml)

    #...........................................................................
    # construct a validator from the tool.pyproj.config table
    config_default = dict()

    for k,v in self.pyproj.config.items():
      if isinstance(v, bool):
        config_default[k] = valid(v, norm_bool)

      elif isinstance(v, scalar_list):
        config_default[k] = restrict(*v)

      else:
        config_default[k] = valid(v, type(v))

    class valid_config(valid_dict):
      _allow_keys = list()
      _default = config_default

    with validating( key = 'config_settings' ):
      _config_settings = valid_config(config_settings or dict())

      self.pyproj.config = _config_settings

    #...........................................................................
    self.build_backend = mapget( self.pptoml,
      'build-system.build-backend',
      "" )

    self.backend_path = mapget( self.pptoml,
      'build-system.backend-path',
      list() )

    #...........................................................................
    # default build requirements
    self.build_requires = self.pptoml.build_system.requires

    #...........................................................................
    # used to create name for binary distribution
    self.is_platlib = bool(self.binary.platlib.copy)

    if self.is_platlib:
      self.binary.compat_tags = platlib_compat_tags()

    #...........................................................................
    self.prep()

    with validating(key = 'project', root = self._pptoml, file = self.pptoml_file):
      self.pkg_info = PkgInfo(
        project = self.project,
        root = root )

    # Update logger once package info is created
    self.logger = self.logger.getChild( f"['{self.pkg_info.name_normed}']" )

  #-----------------------------------------------------------------------------
  @property
  def pptoml(self):
    """pptoml : Parsed and validated pyproject.toml document
    """
    return self._pptoml

  #-----------------------------------------------------------------------------
  @property
  def project(self):
    """:class:`partis.pyproj.pptoml.project`
    """
    return self._pptoml.project

  #-----------------------------------------------------------------------------
  @property
  def pyproj(self):
    """:class:`partis.pyproj.pptoml.pyproj`
    """
    return self._pptoml.tool.pyproj

  #-----------------------------------------------------------------------------
  @property
  def config(self):
    """:class:`partis.pyproj.pptoml.pyproj_config`
    """
    return self._pptoml.tool.pyproj.config

  #-----------------------------------------------------------------------------
  @property
  def meson(self):
    """:class:`partis.pyproj.pptoml.pyproj_meson`
    """
    return self._pptoml.tool.pyproj.meson

  #-----------------------------------------------------------------------------
  @property
  def dist(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist`
    """
    return self._pptoml.tool.pyproj.dist

  #-----------------------------------------------------------------------------
  @property
  def source(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist_source`
    """
    return self._pptoml.tool.pyproj.dist.source

  #-----------------------------------------------------------------------------
  @property
  def binary(self):
    """:class:`partis.pyproj.pptoml.pyproj_dist_binary`
    """
    return self._pptoml.tool.pyproj.dist.binary


  #-----------------------------------------------------------------------------
  @property
  def add_legacy_setup(self):
    """bool
    """
    return self.dist.source.add_legacy_setup

  #-----------------------------------------------------------------------------
  @property
  def build_requires(self):
    """set[:class:`PkgInfoReq`]
    """
    return self._build_requires

  #-----------------------------------------------------------------------------
  @build_requires.setter
  def build_requires(self, reqs):
    self._build_requires = set([ PkgInfoReq(r) for r in reqs ])

  #-----------------------------------------------------------------------------
  def prep_entrypoint( self, name, obj, logger ):

    prep = obj.get( 'prep', None )

    if not prep:
      return None

    prep_name = name

    entry_point = prep.entry
    entry_point_kwargs = prep.kwargs

    try:
      func = load_entrypoint(
        entry_point = entry_point,
        root = self.root )

      logger.info(f"loaded entry-point '{entry_point}'")

    except Exception as e:
      raise EntryPointError(f"failed to load entry-point '{entry_point}'") from e

    try:
      cwd = os.getcwd()

      with validating( file = f"{prep_name} -> {entry_point}(**{entry_point_kwargs})" ):
        func(
          self,
          logger = logger,
          **entry_point_kwargs )

    except Exception as e:
      raise EntryPointError(f"failed to run entry-point '{entry_point}'") from e

    finally:
      os.chdir(cwd)


  #-----------------------------------------------------------------------------
  def prep( self ):
    """Prepares project metadata
    """
    # backup project to detect changes made by prep
    project = deepcopy(self.project)
    dynamic = project.dynamic


    if dynamic and 'prep' not in self.pyproj:
      raise ValidationError(f"tool.pyproj.prep is required to resolve project.dynamic")

    self.prep_entrypoint(
      name = f"tool.pyproj.prep",
      obj = self.pyproj,
      logger = self.logger.getChild( f"prep" ) )

    # NOTE: check that any dynamic meta-data is defined after prep
    for k in dynamic:
      # all dynamic keys should updated by prep
      if k not in self.project or self.project[k] == project[k]:
        warnings.warn(
          f"project.dynamic listed key as dynamic, but not altered in prep: {k}",
          ValidationWarning )

    for k, v in self.project.items():
      if k not in dynamic and (k not in project or project[k] != v):
        # don't allow keys to be added or changed unless they were listed in dynamic
        raise ValidationError(
          f"prep updated key not listed in project.dynamic: {k}" )

    # fields are no longer dynamic
    self.project.dynamic = list()

    # make sure build requirements are still a set of PkgInfoReq
    self.build_requires = self.build_requires

  #-----------------------------------------------------------------------------
  def dist_prep( self ):
    """Prepares project files for a distribution
    """

    self.prep_entrypoint(
      name = f"tool.pyproj.dist.prep",
      obj = self.dist,
      logger = self.logger.getChild( f"dist.prep" ) )


  #-----------------------------------------------------------------------------
  def dist_source_prep( self ):
    """Prepares project files for a source distribution
    """

    self.prep_entrypoint(
      name = f"tool.pyproj.dist.source.prep",
      obj = self.dist.source,
      logger = self.logger.getChild( f"dist.source.prep" ) )

  #-----------------------------------------------------------------------------
  def dist_source_copy( self, *, dist ):
    """Copies prepared files into a source distribution

    Parameters
    ---------
    sdist : :class:`dist_base <partis.pyproj.dist_file.dist_base.dist_base>`
      Builder used to write out source distribution files
    """

    with validating( key = f'tool.pyproj.dist.source' ):
      dist_copy(
        base_path = dist.named_dirs['root'],
        include = self.source.copy,
        ignore = self.dist.ignore + self.source.ignore,
        dist = dist,
        root = self.root,
        logger = self.logger )

      if self.add_legacy_setup:
        with validating( key = f'add_legacy_setup' ):

          self.logger.info(f"generating legacy 'setup.py'")
          legacy_setup_content( self, dist )

  #-----------------------------------------------------------------------------
  def dist_binary_prep( self ):
    """Prepares project files for a binary distribution
    """

    with MesonBuild(
      root = self.root,
      meson = self.meson,
      logger = self.logger.getChild( f"meson" ) ):

      self.prep_entrypoint(
        name = f"tool.pyproj.dist.binary.prep",
        obj = self.binary,
        logger = self.logger.getChild( f"dist.binary.prep" ) )

    self.logger.debug(f"Compatibility tags after dist.binary.prep: {self.binary.compat_tags}")

  #-----------------------------------------------------------------------------
  def dist_binary_copy( self, *, dist ):
    """Copies prepared files into a binary distribution

    Parameters
    ---------
    bdist : :class:`dist_base <partis.pyproj.dist_file.dist_base.dist_base>`
      Builder used to write out binary distribution files
    """


    with validating( key = f'tool.pyproj.dist.binary' ):
      ignore = self.dist.ignore + self.dist.binary.ignore

      dist_copy(
        base_path = dist.named_dirs['root'],
        include = self.binary.copy,
        ignore = ignore,
        dist = dist,
        root = self.root,
        logger = self.logger )

      data_scheme = [
        'data',
        'headers',
        'scripts',
        'purelib',
        'platlib' ]

      for k in data_scheme:
        if k in self.binary:

          dist_data = self.binary[k]

          _include = dist_data.copy
          _ignore = ignore + dist_data.ignore

          with validating( key = k ):
            dist_copy(
              base_path = dist.named_dirs[k],
              include = _include,
              ignore = _ignore,
              dist = dist,
              root = self.root,
              logger = self.logger )



import sys
import os
import os.path as osp
import tempfile
import shutil
import subprocess
import glob

from pytest import (
  mark,
  warns,
  raises )

from partis.pyproj import (
  ValidationError,
  FileOutsideRootError,
  ValidationWarning,
  EntryPointError,
  PyProjBase,
  dist_source_targz,
  dist_binary_wheel )

SKIP_MESON = False

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def uninstall( name, ignore_errors = False ):

  try:
    subprocess.check_call([
      sys.executable,
      '-m',
      'pip',
      'uninstall',
      '-y',
      name ])
  except Exception as e:
    if ignore_errors:
      pass
    else:
      raise e

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def install( name ):
  subprocess.check_call([
    sys.executable,
    '-m',
    'pip',
    'install',
    name ])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def try_dist(
  import_name,
  install_name ):

  # ensure not installed, e.g. from a previous test
  uninstall(
    import_name,
    ignore_errors = True )

  # install built test distribution
  install( install_name )

  # ensure the installation leads to importable module
  __import__( import_name )

  # should be able to succesfully uninstall (don't ignore errors)
  uninstall( import_name )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def try_legacy( name, dist_file ):
  # ensure not installed, e.g. from a previous test
  uninstall(
    name,
    ignore_errors = True )

  with tempfile.TemporaryDirectory() as tmpdir:

    cwd = os.getcwd()

    try:
      os.chdir(tmpdir)

      import tarfile

      with tarfile.open( dist_file ) as fp:

        fp.extractall('.')

      dist_dir = osp.join( tmpdir, osp.basename(dist_file)[:-7] )

      os.chdir(dist_dir)

      subprocess.check_call([
        sys.executable,
        'setup.py',
        'egg_info',
        '-e',
        tmpdir ])

      print(os.listdir(tmpdir))

      egg_info = next(iter(glob.glob(tmpdir + '/*.egg-info')))

      assert osp.isdir(egg_info)

      egg_files = os.listdir(egg_info)

      assert (
        set(egg_files) == set([
          'PKG-INFO',
          'setup_requires.txt',
          'requires.txt',
          'SOURCES.txt',
          'top_level.txt',
          'entry_points.txt',
          'dependency_links.txt',
          'not-zip-safe' ]) )

      subprocess.check_call([
        sys.executable,
        'setup.py',
        'bdist_wheel',
        '-d',
        tmpdir ])

      print(os.listdir(tmpdir))

      wheel_file = next(iter(glob.glob(tmpdir + '/*.whl')))

      try_dist(
        import_name = name,
        install_name = wheel_file )

      subprocess.check_call([
        sys.executable,
        'setup.py',
        'install' ])

      __import__( name )

      # should be able to succesfully uninstall (don't ignore errors)
      uninstall( name )

    finally:
      os.chdir(cwd)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def run_pyproj( name, source = True, binary = True ):

  with tempfile.TemporaryDirectory() as tmpdir:

    outdir = osp.join(tmpdir, 'dist')
    pkg_dir = osp.join( tmpdir, name )

    shutil.copytree(
      osp.join(osp.dirname(osp.abspath(__file__)), name ),
      pkg_dir,
      # some tests require copying symlinks
      symlinks = True )

    cwd = os.getcwd()

    try:
      os.chdir(pkg_dir)

      pyproj = PyProjBase(
        root = pkg_dir )

      pyproj.dist_prep()

      # build and install source dist
      if source:
        pyproj.dist_source_prep()

        with dist_source_targz(
          pkg_info = pyproj.pkg_info,
          outdir = outdir,
          logger = pyproj.logger ) as dist:

          pyproj.dist_source_copy(
            dist = dist )

        try_dist(
          import_name = pyproj.pkg_info.name,
          install_name = dist.outpath )

        if pyproj.add_legacy_setup:
          try_legacy(
            name = pyproj.pkg_info.name,
            dist_file = dist.outpath )

      # build and install binary dist
      if binary:
        pyproj.dist_binary_prep()

        with dist_binary_wheel(
          pkg_info = pyproj.pkg_info,
          compat = pyproj.binary.compat_tags,
          outdir = outdir,
          logger = pyproj.logger ) as dist:

          pyproj.dist_binary_copy(
            dist = dist )

        try_dist(
          import_name = pyproj.pkg_info.name,
          install_name = dist.outpath )

    finally:
      os.chdir(cwd)


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_min():
  run_pyproj('pkg_min')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_base():
  run_pyproj('pkg_base')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_bad_1():
  with raises(EntryPointError):
    # declares non-existent entrypoint
    run_pyproj('pkg_bad_1')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_bad_2():
  with raises(EntryPointError):
    # entrypoint raises exception
    run_pyproj('pkg_bad_2')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_bad_3():
  with raises(ValidationError):
    # declares dynamic but no entry-point
    run_pyproj('pkg_bad_3')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_bad_4():
  with raises(ValidationError):
    # changes meta-data without declaring dynamic
    run_pyproj('pkg_bad_4')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_bad_5():
  with warns(ValidationWarning):
    # declares dynamic but doesn't update
    run_pyproj('pkg_bad_5')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@mark.skipif(SKIP_MESON, reason="")
def test_meson_1():
  run_pyproj('pkg_meson_1')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@mark.skipif(SKIP_MESON, reason="")
def test_meson_2():
  run_pyproj('pkg_meson_2')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TODO: this test is broken
@mark.skipif(SKIP_MESON, reason="")
def test_meson_bad_1():
  with raises(FileOutsideRootError):
    # symlink points outside of root, cannot copy into distro
    run_pyproj(
      'pkg_meson_bad_1',
      source = True,
      binary = False )

  with raises(FileOutsideRootError):
    # symlink points outside of root, cannot use as meson 'prefix'
    run_pyproj(
      'pkg_meson_bad_1',
      source = False,
      binary = True )



import os
import os.path as osp
import io
import warnings
import stat

import tempfile
import shutil
import tarfile

from .dist_base import dist_base

from ..norms import (
  norm_path,
  norm_data,
  norm_mode )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class dist_targz( dist_base ):
  """Builds a tar-file  with gz compression

  Example
  -------

  .. code-block:: python

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:

      import os
      import os.path

      pkg_dir = os.path.join( tmpdir, 'src', 'my_package' )
      out_dir = os.path.join( tmpdir, 'build' )

      os.makedirs( pkg_dir )

      with open( os.path.join( pkg_dir, 'module.py' ), 'w' ) as fp:
        fp.write("print('hello')")

      from partis.pyproj import (
        dist_targz )

      with dist_targz(
        outname = 'my_dist.tar.gz',
        outdir = out_dir ) as dist:

        dist.copytree(
          src = pkg_dir,
          dst = 'my_package' )


  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    outname,
    outdir = None,
    tmpdir = None,
    named_dirs = None,
    logger = None ):

    super().__init__(
      outname = outname,
      outdir = outdir,
      tmpdir = tmpdir,
      named_dirs = named_dirs,
      logger = logger )

    self._fd = None
    self._fp = None
    self._tmp_path = None
    self._tarfile = None

  #-----------------------------------------------------------------------------
  def create_distfile( self ):

    ( self._fd, self._tmp_path ) = tempfile.mkstemp(
      dir = self.tmpdir )

    self._fp = os.fdopen( self._fd, "w+b" )

    self._tarfile = tarfile.open(
      fileobj = self._fp,
      mode = 'w:gz',
      format = tarfile.PAX_FORMAT )


  #-----------------------------------------------------------------------------
  def close_distfile( self ):

    if self._tarfile is not None:

      # close the file
      self._tarfile.close()
      self._tarfile = None

    if self._fp is not None:
      self._fp.close()
      self._fp = None

    if self._fd is not None:
      self._fd = None

  #-----------------------------------------------------------------------------
  def copy_distfile( self ):

    if osp.exists( self.outpath ):
      # overwiting in destination directory
      os.remove( self.outpath )

    if not osp.exists( self.outdir ):
      os.makedirs( self.outdir )

    shutil.copyfile( self._tmp_path, self.outpath )

  #-----------------------------------------------------------------------------
  def remove_distfile( self ):

    # remove temporary file
    os.remove( self._tmp_path )

    self._tmp_path = None


  #-----------------------------------------------------------------------------
  def write( self,
    dst,
    data,
    mode = None,
    record = True ):

    self.assert_open()

    dst = norm_path( dst )

    data = norm_data( data )

    info = tarfile.TarInfo( dst )

    info.size = len(data)
    info.mode = norm_mode( mode )

    self._tarfile.addfile(
      info,
      fileobj = io.BytesIO(data) )

    super().write(
      dst = dst,
      data = data,
      mode = mode,
      record = record )

  #-----------------------------------------------------------------------------
  def finalize( self ): # pragma: no cover
    pass

  #-----------------------------------------------------------------------------
  def exists( self,
    dst ):

    self.assert_open()

    try:
      self._tarfile.getmember(dst)
      return True
    except KeyError as e:
      return False




import io
import re
import pathlib
from pytest import (
  raises )

from email.utils import formataddr

from partis.pyproj import (
  scalar,
  scalar_list,
  empty_str,
  nonempty_str,
  str_list,
  nonempty_str_list,
  norm_bool,
  CompatibilityTags,
  ValidationError,
  PEPValidationError,
  valid_type,
  valid_keys,
  as_list,
  mapget,
  norm_printable,
  valid_dist_name,
  norm_dist_name,
  norm_dist_filename,
  join_dist_filename,
  norm_dist_version,
  norm_dist_author,
  norm_dist_classifier,
  norm_dist_keyword,
  norm_dist_url,
  norm_dist_extra,
  norm_dist_build,
  dist_build,
  norm_dist_compat,
  join_dist_compat,
  compress_dist_compat,
  norm_data,
  norm_py_identifier,
  norm_entry_point_group,
  norm_entry_point_name,
  norm_entry_point_ref,
  norm_path,
  norm_path_to_os,
  norm_mode,
  norm_zip_external_attr,
  b64_nopad,
  hash_sha256,
  email_encode_items )

from partis.pyproj._nonprintable import (
  _gen_nonprintable,
  gen_nonprintable )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_scalars():
  #.............................................................................
  xs = [False, 0, 0.0, '000', True, 1, 1.0, '111', '']

  for x in xs:
    assert scalar(x) is x

  assert scalar_list(xs) == xs

  ys = [ [], [1,2,3], {}, {1:1}, set() ]

  for y in ys:
    with raises( ValidationError ):
      scalar(y)

  with raises( ValidationError ):
    scalar_list(ys)

  #.............................................................................
  ts = [1, 1.0, True, 'true', 'True', 'yes', 'y', 'enable', 'enabled']
  fs = [0, 0.0, False, 'false', 'False', 'no', 'n', 'disable', 'disabled']

  assert all( norm_bool(t) for t in ts)
  assert not any( norm_bool(f) for f in fs )

  with raises( ValidationError ):
    norm_bool(11)

  with raises( ValidationError ):
    norm_bool(1.1)

  with raises( ValidationError ):
    norm_bool('')

  with raises( ValidationError ):
    norm_bool('1')

  #.............................................................................
  assert empty_str('') == ''

  with raises( ValidationError ):
    empty_str('123')

  with raises( ValidationError ):
    empty_str(123)


  assert nonempty_str('123') == '123'
  assert nonempty_str(123) == '123'

  with raises( ValidationError ):
    nonempty_str('')

  #.............................................................................
  zs = ['1', '2', '3']
  assert str_list(zs) == zs

  qs = [1, 2, 3]
  assert str_list(qs) == zs

  #.............................................................................
  assert nonempty_str_list(zs) == zs

  with raises( ValidationError ):
    nonempty_str_list(['', '', '123'])


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_as_list():
  assert as_list(None) == [None]
  assert as_list('a') == ['a']
  assert as_list(['a', 'b']) == ['a', 'b']
  assert as_list({'a': 'b'}) == ['a']

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_printable():

  regex = gen_nonprintable()
  ns, test = _gen_nonprintable()

  test = norm_printable(norm_printable)

  # since norm_printable considers \t and \n to be printable, but isprintable does not
  test = re.sub(r'[\t\n]', '', test)

  assert test.isprintable()


  assert norm_printable(None) == ''
  assert norm_printable("") == ''
  assert norm_printable("hello\t\tfoo bar\ngoodbye\n\n") == "hello\t\tfoo bar\ngoodbye"
  assert norm_printable("\U0001EE78") == ''
  assert norm_printable("f\ubaaar") == "f몪r"

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valid_dist_name():
  valid_names = [
    'xyz',
    '\txyz\n',
    'x_y_z',
    'x1_y2',
    'x.y.z',
    'x-y-z' ]

  invalid_names = [
    '_123',
    '-1x',
    '.x1'
    'x y z',
    '']

  for name in valid_names:
    assert name.strip() == valid_dist_name(name)

  for name in invalid_names:
    print(name)
    with raises( PEPValidationError ):
      valid_dist_name(name)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_name():
  names = [
    ('  x.-__.--__Y0.z \n', 'x-y0-z') ]

  for name, val in names:
    assert val == norm_dist_name(name)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_filename():
  names = [
    ('  x.-__.--__y0.z \n', 'x_y0_z') ]

  for name, val in names:
    assert val == norm_dist_filename(norm_dist_name(name))

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_join_dist_filename():
  assert 'w_x-y-z' == join_dist_filename(['w--x','y','','','z'])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_version():
  valid = [
    '1',
    '1.2',
    '1.2.3',
    ' 1.2.3\n',
    '1.2.3a0',
    '1.2.3b12',
    '1.2.3rc123',
    '1.2.3.post0']

  invalid = [
    'xyz']

  for x in valid:
    assert x.strip() == norm_dist_version(x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_version(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_author():
  valid = [
    (('', ''), ('', '')),
    (('x', ''), ('x', '')),
    (('', 'y@z.com'), ('', 'y@z.com')),
    (('x', 'y@z.com'), ('', formataddr( ('x', 'y@z.com') )) ) ]

  invalid = [
    ('', 'xyz>'),
    ('f\ubaaar', ''),
    ('a,', ''),
    ('', 'xyz')  ]

  for x, y in valid:
    assert y == norm_dist_author(*x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_author(*x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_classifier():
  valid = [
    ( 'x   \n:: y  ', 'x :: y' ) ]

  invalid = [
    "%",
    "*",
    "asd :: *" ]

  for x, y in valid:
    assert y == norm_dist_classifier(x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_classifier(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_keyword():
  valid = [
    "asd " ]

  invalid = [
    "asd bfr",
    "asd, bfr" ]

  for x in valid:
    assert x.strip() == norm_dist_keyword(x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_keyword(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_url():
  valid = [
    (('xyz','http://xyz.com/123'), ('xyz','http://xyz.com/123')) ]

  invalid = [
    ('', ''),
    ('a,', ''),
    ('', '(*&(*&))')  ]

  for x, y in valid:
    assert y == norm_dist_url(*x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_url(*x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_extra():
  valid = [
    "asd " ]

  invalid = [
    "asd bfr" ]

  for x in valid:
    assert x.strip() == norm_dist_extra(x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_extra(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_build():
  valid = [
    "1A" ]

  invalid = [
    "a1" ]

  for x in valid:
    assert x.lower() == norm_dist_build(x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_build(x)


  assert dist_build() == ''
  assert dist_build(1) == '1'
  assert dist_build(build_tag = 'asd') == '0_asd'
  assert dist_build(123, 'asd') == '123_asd'

  with raises( ValueError ):
    dist_build('qwe', 'asd')

  with raises( PEPValidationError ):
    dist_build(123, 'asd-test')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_dist_compat():
  valid = [
    ( 'py3', 'none', 'any' ) ]

  invalid = [
    ( '', 'none', 'any' ),
    ( 'py3', '', 'any' ),
    ( 'py3', 'none', '' ) ]

  for x in valid:
    assert x == norm_dist_compat(*x)

  for x in invalid:
    print(x)
    with raises( PEPValidationError ):
      norm_dist_compat(*x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_join_dist_compat():
  assert 'x.y.z' == join_dist_compat(['z','x','x','y'])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_compress_dist_compat():
  assert ( "py2.py3", "cp3.none", "any.linux" ) == compress_dist_compat([
    ( 'py3', 'cp3', 'linux' ),
    ( 'py2', 'none', 'any' ) ])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_data():
  assert norm_data("asd") == "asd".encode('utf-8')
  assert norm_data(b"asd") == b"asd"
  assert norm_data(io.BytesIO(b"asd")) == b"asd"

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_py_identifier():
  valid = [
    "asd",
    " asd\n",
    "a1" ]

  invalid = [
    "1a",
    "import" ]

  for x in valid:
    assert x.strip() == norm_py_identifier(x)

  for x in invalid:
    print(x)
    with raises( ValidationError ):
      norm_py_identifier(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_entry_point_group():
  valid = [
    "a.b.c",
    " a.b.c\n" ]

  invalid = [
    "a b c" ]

  for x in valid:
    assert x.strip() == norm_entry_point_group(x)

  for x in invalid:
    print(x)
    with raises( ValidationError ):
      norm_entry_point_group(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_entry_point_name():
  valid = [
    "a.b.c",
    " a.b.c\n" ]

  invalid = [
    "%$" ]

  for x in valid:
    assert x.strip() == norm_entry_point_name(x)

  for x in invalid:
    print(x)
    with raises( ValidationError ):
      norm_entry_point_name(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_entry_point_ref():
  valid = [
    ("a.b.c", "a.b.c"),
    ("a.b.c : xyz ", "a.b.c:xyz") ]

  invalid = [
    ":asd",
    "a.b.c ; xyz ",
    "a.1b.c" ]

  for x, y in valid:
    assert norm_entry_point_ref(x) == y

  for x in invalid:
    print(x)
    with raises( ValidationError ):
      norm_entry_point_ref(x)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_path():

  valid = [
    ("a/b/c", "a/b/c"),
    (r"a\b\c", "a/b/c") ]

  invalid = [
    "/asd",
    "asd/a b c/xyz",
    "a/b/../..",
    "../"]

  for x, y in valid:
    assert norm_path(x) == y

  for x in invalid:
    print(x)
    with raises( ValidationError ):
      norm_path(x)


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_path_to_os():

  assert norm_path_to_os(__file__) == __file__

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_mode():
  assert norm_mode() == 0o644
  assert norm_mode('1') == 0o644

  assert norm_mode(0o755) == 0o755
  assert norm_mode(0o744) == 0o755

  assert norm_mode(0o655) == 0o644
  assert norm_mode(0o644) == 0o644

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_norm_zip_external_attr():
  assert norm_zip_external_attr(0o644) == 0o644 << 16

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_b64_nopad():
  assert b64_nopad(b'data') == 'ZGF0YQ'

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_hash_sha256():
  data = b'data'

  assert hash_sha256(data) == hash_sha256(io.BytesIO(data))

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_email_encode_items():

  b = email_encode_items(
    headers = [
      ('a', 'b'),
      ('c', 'd') ],
    payload = "hello world" )

  c = b.decode('ascii')

  assert c == "a: b\nc: d\n\nhello world"



import os
import os.path as osp
import sys
import importlib
from pathlib import Path

from .norms import (
  norm_path_to_os )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class EntryPointError(ValueError):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def module_name_from_path( path, root ):
  """Generates an importable module name from a file system path

  Parameters
  ----------
  path : str
    Path to the module directory relative to 'root'
  root : str
    Base path from which the module will be imported
  """
  path = Path( path )
  root = Path( root )

  # path = path.with_suffix("")

  relative_path = osp.relpath( path, start = root )

  path_parts = Path(relative_path).parts

  if len(path_parts) == 0:
    raise EntryPointError("Empty module name")

  return ".".join(path_parts)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def load_module( path, root ):

  if not osp.isdir(path):
    raise EntryPointError(f"Not a directory: {path}")

  init_file = osp.join( path, "__init__.py" )

  if not osp.isfile(init_file):
    raise EntryPointError(f"Not a module: {init_file}")

  module_name = module_name_from_path( path = path, root = root )

  spec = importlib.util.spec_from_file_location(
    name = module_name,
    location = str(init_file) )

  mod = importlib.util.module_from_spec( spec )
  sys.modules[ spec.name ] = mod

  spec.loader.exec_module( mod )

  return mod

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def load_entrypoint( entry_point, root ):

  mod_name, attr_name = entry_point.split(':')

  mod_name = mod_name.strip()
  attr_name = attr_name.strip()

  mod = load_module(
    path = osp.join( root, norm_path_to_os( mod_name.replace('.', '/') ) ),
    root = root )

  if not hasattr( mod, attr_name ):
    raise EntryPointError(
      f"'{mod_name}' :'{attr_name}'" )

  func = getattr( mod, attr_name )

  return func



import os
import os.path as osp
import tempfile
import shutil

from pytest import (
  raises )

from partis.pyproj.load_module import (
  module_name_from_path,
  load_module,
  load_entrypoint )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_name():

  assert module_name_from_path(osp.join('a','b','c'), '') == "a.b.c"
  assert module_name_from_path(osp.join('a', 'b', 'c', 'd'), osp.join('a','b')) == "c.d"

  with raises(ValueError):
    # no module name
    module_name_from_path("", "")

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_load_module():
  root = osp.dirname(osp.abspath(__file__))

  mod = load_module( osp.join(root, "pkg_base", "pkgaux" ), root )

  print(dir(mod))
  assert callable(mod.dist_prep)

  with raises( ValueError ):
    load_module( osp.join(root, "not_a_directory" ), root )

  with raises( ValueError ):
    # not a module
    load_module( osp.join(root, "pkg_base" ), root )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_load_entrypoint():
  root = osp.dirname(osp.abspath(__file__))

  f = load_entrypoint('pkg_base.pkgaux:dist_prep', root)

  assert callable(f)

  with raises( ValueError ):
    load_entrypoint('pkg_base.pkgaux:not_an_attr', root)



"""Convenience CLI for running the tests

$ python -m test
"""

import os
import os.path as osp
import subprocess
import argparse
from argparse import RawTextHelpFormatter

parser = argparse.ArgumentParser(
  description = __doc__,
  formatter_class = RawTextHelpFormatter )

parser.add_argument( "-s", "--session",
  type = str,
  default = None,
  help = "session to run with nox (--list to see available sessions)" )

parser.add_argument( "--list",
  action = 'store_true',
  help = "list available nox sessions" )

args = parser.parse_args( )

subprocess.check_call([
  'nox',
  '-f',
  osp.join( osp.dirname(__file__), 'noxfile.py' ),
  *( ['-s', args.session ] if args.session else [] ),
  *( ['--list', ] if args.list else [] ) ])



import os
import os.path as osp
import tempfile
import shutil
import logging

from pytest import (
  raises )

from partis.pyproj.backend import (
  UnsupportedOperation,
  backend_init,
  get_requires_for_build_sdist,
  build_sdist,
  get_requires_for_build_wheel,
  prepare_metadata_for_build_wheel,
  build_wheel )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_backend_basic():
  root = osp.join(osp.dirname(osp.abspath(__file__)), 'pkg_base' )

  a = backend_init(
    root = root )

  b = backend_init(
    root = root,
    logger = logging.getLogger( __name__ )  )

  with raises(FileNotFoundError):
    c = backend_init(
      root = osp.dirname(osp.abspath(__file__)) )


  cwd = os.getcwd()

  try:
    os.chdir( root )

    # currently does not return any additional requirements
    assert get_requires_for_build_sdist() == list()
    assert get_requires_for_build_wheel() == list()

    with tempfile.TemporaryDirectory() as tmpdir:

      name =  build_sdist( dist_directory = tmpdir )
      assert osp.exists( osp.join( tmpdir, name ) )

      name =  prepare_metadata_for_build_wheel( metadata_directory = tmpdir )
      assert osp.exists( osp.join( tmpdir, name) )

      name =  build_wheel( wheel_directory = tmpdir )
      assert osp.exists( osp.join( tmpdir, name) )

  finally:
    os.chdir( cwd )



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# This is the config file for running the nox testing automation tool
# This will run pytest and generate a combined coverage report for all runs

import nox
import sys
import os
import os.path as osp
import re
import tomli
import itertools
import subprocess

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Config
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
test_dir = osp.dirname(osp.abspath(__file__))
root_dir = osp.join( test_dir, os.pardir )
source_dir =  osp.join( root_dir, 'src' )
pptoml_file = osp.join( root_dir, 'pyproject.toml' )
sitcustom_dir = osp.join( test_dir, 'sitecustom' )
flake8_file = osp.join( test_dir, '.flake8' )

with open( pptoml_file, 'r' ) as fp:
  pptoml = tomli.loads( fp.read() )

opt_deps = pptoml['project']['optional-dependencies']

test_deps = opt_deps['test']
test_deps += [sitcustom_dir]

cov_deps = opt_deps['cov']
lint_deps = opt_deps['lint']

ppnox = pptoml['tool']['noxfile']
python_versions = ppnox['python']
nox.options.envdir = osp.join( root_dir, ppnox['envdir'] )
nox.options.default_venv_backen = 'venv'


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Sessions
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@nox.session()
def prepare(session):
  session.chdir('..')

  session.install(*cov_deps)

  session.env['COVERAGE_RCFILE'] = pptoml_file

  session.run('coverage', 'erase')

  session.run(
    'python',
    '-m',
    'make_dists',
    '--no-doc' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# successivly build and install sdist/wheel, run tests as individual sub-projects
@nox.session(
  python = python_versions )
def test( session ):

  session.chdir('..')

  # coverage data for this sessions
  name = re.sub(r"[^A-Za-z0-9]+", "", session.name )
  session.env['COVERAGE_FILE'] = osp.join( root_dir, 'tmp', f'.coverage.{name}' )

  # global coverage config
  session.env['COVERAGE_RCFILE'] = pptoml_file

  # initialize in subprocess coverage hook
  session.env['COVERAGE_PROCESS_START'] = pptoml_file

  # this is so the sdist/wheel can be found by pip to resolve inter-package deps.
  session.env['PIP_FIND_LINKS'] = osp.join(root_dir, 'dist')

  # remove from the pip cache to prevent using a previously installed distro.
  session.run(
    'python3',
    '-m',
    'pip',
    'cache',
    'remove',
    "'partis-pyproj'" )

  session.install(
    *test_deps,
    'partis-pyproj' )

  session.run(
    'pytest',
    test_dir )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@nox.session( venv_backend = 'venv' )
def report(session):
  session.chdir('..')

  session.install(*cov_deps)

  session.env['COVERAGE_RCFILE'] = pptoml_file

  session.run('coverage', 'combine')
  session.run('coverage', 'report')
  session.run('coverage', 'html')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# @nox.session()
# def lint(session):
#   session.chdir('..')
#
#   session.install(*lint_deps)
#
#   session.run(
#     'python3',
#     '-m',
#     'pyflakes',
#     source_dir )



"""Usage of `setup.py` is deprecated, and is supplied only for legacy installation.
"""
import sys
import os
import os.path as osp
import importlib
import logging
import argparse
import subprocess
import tempfile
from argparse import RawTextHelpFormatter
logger = logging.getLogger(__name__)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def egg_info( args ):

  logger.warning(
    "running legacy 'setup.py egg_info'" )

  dir = osp.join( args.egg_base, EGG_INFO_NAME )

  if not osp.exists( dir ):
    os.mkdir( dir )

  with open( osp.join( dir, 'PKG-INFO' ), 'wb' ) as fp:
    fp.write( PKG_INFO )

  with open( osp.join( dir, 'setup_requires.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'requires.txt' ), 'wb' ) as fp:
    fp.write( REQUIRES )

  with open( osp.join( dir, 'SOURCES.txt' ), 'wb' ) as fp:
    fp.write( SOURCES )

  with open( osp.join( dir, 'top_level.txt' ), 'wb' ) as fp:
    fp.write( TOP_LEVEL )

  with open( osp.join( dir, 'entry_points.txt' ), 'wb' ) as fp:
    fp.write( ENTRY_POINTS )

  with open( osp.join( dir, 'dependency_links.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'not-zip-safe' ), 'wb' ) as fp:
    fp.write( b'' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def bdist_wheel( args ):

  logger.warning(
    "running legacy 'setup.py bdist_wheel'" )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  backend.build_wheel(
    wheel_directory = args.dist_dir or args.bdist_dir or '.' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def install( args ):

  logger.warning(
    "running legacy 'setup.py install'" )

  reqs = [ f"{r}" for r in build_requires ]

  subprocess.check_call([
    sys.executable,
    '-m',
    'pip',
    'install',
    *reqs ] )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  with tempfile.TemporaryDirectory() as tmpdir:
    wheel_name = backend.build_wheel(
      wheel_directory = tmpdir )

    subprocess.check_call([
      sys.executable,
      '-m',
      'pip',
      'install',
      osp.join(tmpdir, wheel_name) ])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dummy( args ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def main():

  logging.basicConfig(
    level = logging.INFO,
    format = "{name}:{levelname}: {message}",
    style = "{" )


  logger.warning(
    "'setup.py' is deprecated, limited support for legacy installs. Upgrade pip." )

  parser = argparse.ArgumentParser(
    description = __doc__,
    formatter_class = RawTextHelpFormatter )

  subparsers = parser.add_subparsers()

  #.............................................................................
  egg_info_parser = subparsers.add_parser( 'egg_info' )

  egg_info_parser.set_defaults( func = egg_info )

  egg_info_parser.add_argument( "-e", "--egg-base",
    type = str,
    default = '.' )

  #.............................................................................
  bdist_wheel_parser = subparsers.add_parser( 'bdist_wheel' )

  bdist_wheel_parser.set_defaults( func = bdist_wheel )

  bdist_wheel_parser.add_argument( "-b", "--bdist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "-d", "--dist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "--python-tag",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--plat-name",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--py-limited-api",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--build-number",
    type = str,
    default = None )

  #.............................................................................
  install_parser = subparsers.add_parser( 'install' )

  install_parser.set_defaults( func = install )

  install_parser.add_argument( "--record",
    type = str,
    default = None )

  install_parser.add_argument( "--install-headers",
    type = str,
    default = None )

  install_parser.add_argument( "--compile",
    action='store_true' )

  install_parser.add_argument( "--single-version-externally-managed",
    action='store_true' )

  #.............................................................................
  clean_parser = subparsers.add_parser( 'clean' )

  clean_parser.set_defaults( func = dummy )

  clean_parser.add_argument( "-a", "--all",
    action='store_true' )

  args = parser.parse_args( )

  args.func( args )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# NOTE: these are templated literal values substituded by the backend when
# building the source distribution

build_backend = 'pyproj.backend'
backend_path = ['src']
build_requires = ['tomli>=1.2.3', 'wheel', 'packaging==21.3', 'importlib_metadata; python_version < "3.8"']

EGG_INFO_NAME = 'partis-pyproj.egg-info'

PKG_INFO = b'Metadata-Version: 2.1\nName: partis-pyproj\nVersion: 0.0.6\nRequires-Python: >=3.6.2\nAuthor-email: "Nanohmics Inc." <software.support@nanohmics.com>\nMaintainer-email: "Nanohmics Inc." <software.support@nanohmics.com>\nSummary: Minimal set of Python project utilities (PEP-517/621)\nLicense-File: LICENSE.txt\nClassifier: Programming Language :: Python :: 3\nClassifier: Programming Language :: Python\nClassifier: Operating System :: POSIX :: Linux\nClassifier: Development Status :: 4 - Beta\nClassifier: Operating System :: Microsoft :: Windows\nClassifier: License :: OSI Approved :: BSD License\nClassifier: Topic :: Software Development :: Build Tools\nClassifier: Intended Audience :: Developers\nProvides-Extra: meson\nProvides-Extra: doc\nProvides-Extra: test\nProvides-Extra: cov\nProvides-Extra: lint\nRequires-Dist: wheel\nRequires-Dist: packaging==21.3\nRequires-Dist: tomli>=1.2.3\nRequires-Dist: importlib_metadata; python_version < "3.8"\nRequires-Dist: ninja>=1.10.2.3; extra == "meson"\nRequires-Dist: meson>=0.61.3; extra == "meson"\nRequires-Dist: partis-utils[sphinx]==0.0.5; extra == "doc"\nRequires-Dist: Cython>=0.29.18; extra == "test"\nRequires-Dist: pytest-cov>=3.0.0; extra == "test"\nRequires-Dist: pip>=18.1; extra == "test"\nRequires-Dist: nox>=2021.10.1; extra == "test"\nRequires-Dist: tomli>=1.2.3; extra == "test"\nRequires-Dist: coverage[toml]>=6.2; extra == "test"\nRequires-Dist: numpy; extra == "test"\nRequires-Dist: pytest_mock>=3.6.1; extra == "test"\nRequires-Dist: pytest>=6.2.5; extra == "test"\nRequires-Dist: ninja>=1.10.2.3; extra == "test"\nRequires-Dist: build>=0.7.0; extra == "test"\nRequires-Dist: meson>=0.61.3; extra == "test"\nRequires-Dist: coverage[toml]>=6.2; extra == "cov"\nRequires-Dist: pyflakes==2.4.0; extra == "lint"\nDescription-Content-Type: text/x-rst\n\nThe ``partis.pyproj`` package aims to be very simple and\ntransparent implementation of a PEP-517 build back-end.\n\nhttps://nanohmics.bitbucket.io/doc/partis/pyproj'

REQUIRES = b'wheel\npackaging==21.3\ntomli>=1.2.3\nimportlib_metadata; python_version < "3.8"\nninja>=1.10.2.3; extra == "meson"\nmeson>=0.61.3; extra == "meson"\npartis-utils[sphinx]==0.0.5; extra == "doc"\nCython>=0.29.18; extra == "test"\npytest-cov>=3.0.0; extra == "test"\npip>=18.1; extra == "test"\nnox>=2021.10.1; extra == "test"\ntomli>=1.2.3; extra == "test"\ncoverage[toml]>=6.2; extra == "test"\nnumpy; extra == "test"\npytest_mock>=3.6.1; extra == "test"\npytest>=6.2.5; extra == "test"\nninja>=1.10.2.3; extra == "test"\nbuild>=0.7.0; extra == "test"\nmeson>=0.61.3; extra == "test"\ncoverage[toml]>=6.2; extra == "cov"\npyflakes==2.4.0; extra == "lint"'

SOURCES = b'partis_pyproj-0.0.6/src/pyproj/__init__.py\npartis_pyproj-0.0.6/src/pyproj/norms.py\npartis_pyproj-0.0.6/src/pyproj/pep.py\npartis_pyproj-0.0.6/src/pyproj/load_module.py\npartis_pyproj-0.0.6/src/pyproj/_legacy_setup.py\npartis_pyproj-0.0.6/src/pyproj/_nonprintable.py\npartis_pyproj-0.0.6/src/pyproj/validate.py\npartis_pyproj-0.0.6/src/pyproj/pkginfo.py\npartis_pyproj-0.0.6/src/pyproj/meson.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_targz.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/__init__.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_zip.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_binary.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_base.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_source.py\npartis_pyproj-0.0.6/src/pyproj/dist_file/dist_copy.py\npartis_pyproj-0.0.6/src/pyproj/legacy.py\npartis_pyproj-0.0.6/src/pyproj/backend.py\npartis_pyproj-0.0.6/src/pyproj/pyproj.py\npartis_pyproj-0.0.6/src/pyproj/pptoml.py\npartis_pyproj-0.0.6/doc/conf.py\npartis_pyproj-0.0.6/doc/__init__.py\npartis_pyproj-0.0.6/doc/index.rst\npartis_pyproj-0.0.6/doc/src/pptoml.rst\npartis_pyproj-0.0.6/doc/src/index.rst\npartis_pyproj-0.0.6/doc/src/backend.rst\npartis_pyproj-0.0.6/doc/src/validate.rst\npartis_pyproj-0.0.6/doc/src/pkginfo.rst\npartis_pyproj-0.0.6/doc/src/pyproj.rst\npartis_pyproj-0.0.6/doc/src/pep.rst\npartis_pyproj-0.0.6/doc/src/norms.rst\npartis_pyproj-0.0.6/doc/src/dist_file.rst\npartis_pyproj-0.0.6/doc/__main__.py\npartis_pyproj-0.0.6/doc/glossary.rst\npartis_pyproj-0.0.6/test/pkg_bad_5/src/test_pkg/sub_mod/sub_sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_bad_5/src/test_pkg/sub_mod/sub_sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_bad_5/src/test_pkg/pure_mod/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_bad_5/pkgaux/__init__.py\npartis_pyproj-0.0.6/test/pkg_bad_5/pyproject.toml\npartis_pyproj-0.0.6/test/__init__.py\npartis_pyproj-0.0.6/test/pkg_meson_2/src/test_pkg/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_meson_2/src/test_pkg/plat_mod.pyx\npartis_pyproj-0.0.6/test/pkg_meson_2/meson.build\npartis_pyproj-0.0.6/test/pkg_meson_2/meson_options.txt\npartis_pyproj-0.0.6/test/pkg_meson_2/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_bad_2/src/test_pkg/sub_mod/sub_sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_bad_2/src/test_pkg/sub_mod/sub_sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_bad_2/src/test_pkg/pure_mod/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_bad_2/pkgaux/__init__.py\npartis_pyproj-0.0.6/test/pkg_bad_2/pyproject.toml\npartis_pyproj-0.0.6/test/test_02_dist.py\npartis_pyproj-0.0.6/test/__main__.py\npartis_pyproj-0.0.6/test/test_04_pkginfo.py\npartis_pyproj-0.0.6/test/test_06_backend.py\npartis_pyproj-0.0.6/test/sitecustom/partis-sitecustom.pth\npartis_pyproj-0.0.6/test/sitecustom/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/bad_link/src/test_pkg/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/bad_link/src/test_pkg/plat_mod.pyx\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/bad_link/meson.build\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/bad_link/meson_options.txt\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/bad_link/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/src/test_pkg/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/src/test_pkg/plat_mod.pyx\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/meson.build\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/meson_options.txt\npartis_pyproj-0.0.6/test/pkg_meson_bad_1/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_min/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_min/pyproject.toml\npartis_pyproj-0.0.6/test/noxfile.py\npartis_pyproj-0.0.6/test/test_00_validate.py\npartis_pyproj-0.0.6/test/test_01_norms.py\npartis_pyproj-0.0.6/test/pkg_base/src/test_pkg/sub_mod/sub_sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_base/src/test_pkg/sub_mod/sub_sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_base/src/test_pkg/sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_base/src/test_pkg/sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_base/src/test_pkg/pure_mod/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_base/pkgaux/__init__.py\npartis_pyproj-0.0.6/test/pkg_base/pyproject.toml\npartis_pyproj-0.0.6/test/test_03_load_module.py\npartis_pyproj-0.0.6/test/test_05_pyproj.py\npartis_pyproj-0.0.6/test/pkg_bad_3/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_meson_1/src/test_pkg/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_meson_1/src/test_pkg/plat_mod.pyx\npartis_pyproj-0.0.6/test/pkg_meson_1/meson.build\npartis_pyproj-0.0.6/test/pkg_meson_1/meson_options.txt\npartis_pyproj-0.0.6/test/pkg_meson_1/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_bad_4/src/test_pkg/sub_mod/sub_sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_bad_4/src/test_pkg/sub_mod/sub_sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_bad_4/src/test_pkg/pure_mod/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_bad_4/pkgaux/__init__.py\npartis_pyproj-0.0.6/test/pkg_bad_4/pyproject.toml\npartis_pyproj-0.0.6/test/pkg_bad_1/src/test_pkg/sub_mod/sub_sub_mod/good_file.py\npartis_pyproj-0.0.6/test/pkg_bad_1/src/test_pkg/sub_mod/sub_sub_mod/bad_file.py\npartis_pyproj-0.0.6/test/pkg_bad_1/src/test_pkg/pure_mod/pure_mod.py\npartis_pyproj-0.0.6/test/pkg_bad_1/pyproject.toml\npartis_pyproj-0.0.6/pyproject.toml\npartis_pyproj-0.0.6/LICENSE.txt\npartis_pyproj-0.0.6/README.rst'

TOP_LEVEL = b''

ENTRY_POINTS = b''

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

if __name__ == "__main__":
  exit( main() )



import os
import os.path as osp
import tempfile
import shutil
import subprocess

from .validate import (
  validating,
  ValidationError,
  ValidPathError,
  FileOutsideRootError )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def meson_option_arg(k, v):
  """Convert python key-value pair to meson ``-Dkey=value`` option
  """
  if isinstance(v, bool):
    v = ({True: 'true', False: 'false'})[v]

  return f'-D{k}={v}'

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MesonBuild:
  """Run meson setup, compile, install commands

  Parameters
  ----------
  root : str
    Path to root project directory
  meson : :class:`pyproj_meson <partis.pyproj.pptoml.pyproj_meson>`
  logger : logging.Logger
  """
  #-----------------------------------------------------------------------------
  def __init__(self, root, meson, logger):
    self.root = root
    self.meson = meson
    self.logger = logger
    self.meson_paths = dict(
      src_dir = None,
      build_dir = None,
      prefix = None )

  #-----------------------------------------------------------------------------
  def __enter__(self):

    if not self.meson.compile:
      return

    if not shutil.which('meson'):
      raise ValueError(f"The 'meson' program not found.")

    if not shutil.which('ninja'):
      raise ValueError(f"The 'ninja' program not found.")

    # check paths
    for k in ['src_dir', 'build_dir', 'prefix']:
      with validating(key = f"tool.pyproj.meson.{k}"):

        rel_path = self.meson[k]

        abs_path = osp.realpath( osp.join(
          self.root,
          rel_path ) )

        if osp.commonpath([self.root, abs_path]) != self.root:
          raise FileOutsideRootError(
            f"Must be within project root directory:"
            f"\n  file = \"{abs_path}\"\n  root = \"{self.root}\"")

        self.meson_paths[k] = abs_path

    src_dir = self.meson_paths['src_dir']
    build_dir = self.meson_paths['build_dir']
    prefix = self.meson_paths['prefix']

    with validating(key = f"tool.pyproj.meson.src_dir"):
      if not osp.exists(src_dir):
        raise ValidPathError(f"Source directory not found: {src_dir}")

    with validating(key = f"tool.pyproj.meson"):
      if osp.commonpath([build_dir, prefix]) == build_dir:
        raise ValidPathError(f"'prefix' cannot be inside 'build_dir': {build_dir}")

    for k in ['build_dir', 'prefix']:
      with validating(key = f"tool.pyproj.meson.{k}"):
        dir = self.meson_paths[k]

        if dir == self.root:
          raise ValidPathError(f"'{k}' cannot be root directory: {dir}")

        if not osp.exists(dir):
          os.makedirs(dir)

    self.logger.info(f"Running meson build")
    self.logger.info(f"Meson build dir: {build_dir}")
    self.logger.info(f"Meson prefix: {prefix}")

    # TODO: ensure any paths in setup_args are normalized
    if not ( osp.exists(build_dir) and os.listdir(build_dir) ):
      # only run setup if the build directory does not already exist
      setup_args = [
        'meson',
        'setup',
        *self.meson.setup_args,
        '--prefix',
        prefix,
        *[ meson_option_arg(k,v) for k,v in self.meson.options.items() ],
        build_dir,
        self.meson_paths['src_dir'] ]

    elif not self.meson.build_clean:
      # only re-compile if the build directory should be 'clean'
      setup_args = list()

    else:
      raise ValidPathError(
        f"'build_dir' is not empty, remove manually if this is intended or set 'build_clean = false': {build_dir}")

    compile_args = [
      'meson',
      'compile',
      *self.meson.compile_args,
      '-C',
      build_dir ]

    install_args = [
      'meson',
      'install',
      *self.meson.install_args,
      '--no-rebuild',
      '-C',
      build_dir ]


    try:

      if setup_args:
        self.logger.debug(' '.join(setup_args))
        subprocess.check_call(setup_args)

      self.logger.debug(' '.join(compile_args))

      subprocess.check_call(compile_args)

      self.logger.debug(' '.join(install_args))

      subprocess.check_call(install_args)

    except:

      if self.meson.build_clean and osp.exists(build_dir):
        self.logger.info(f"Removing Meson build dir")
        shutil.rmtree(build_dir)

      raise

  #-----------------------------------------------------------------------------
  def __exit__(self, type, value, traceback):
    build_dir = self.meson_paths['build_dir']

    if build_dir is not None and osp.exists(build_dir) and self.meson.build_clean:
      self.logger.info(f"Removing Meson build dir")
      shutil.rmtree(build_dir)

    # do not handle any exceptions here
    return False



import os
import os.path as osp
import tempfile
import shutil

from pytest import (
  raises )

from partis.pyproj import (
  PkgInfo,
  dist_base,
  dist_targz,
  dist_zip,
  dist_source_targz,
  dist_source_dummy,
  dist_binary_wheel )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_base():

  class dist_dummy( dist_base ):
    #-----------------------------------------------------------------------------
    def create_distfile( self ):
      # cause error while opening
      raise NotImplementedError('')

    #-----------------------------------------------------------------------------
    def close_distfile( self ):
      # no error so 'close' will be successful
      pass

    #-----------------------------------------------------------------------------
    def copy_distfile( self ):
      raise NotImplementedError('')

    #-----------------------------------------------------------------------------
    def remove_distfile( self ):
      # no error so 'close' will be successful
      pass

    #-----------------------------------------------------------------------------
    def finalize( self ):
      raise NotImplementedError('')

  dist = dist_dummy('asdasd')

  assert not dist.exists('stuff')

  with raises(NotImplementedError):

    dist.open()


  class dist_dummy2( dist_base ):
    #-----------------------------------------------------------------------------
    def exists( self, dst ):
      return True

    #-----------------------------------------------------------------------------
    def create_distfile( self ):
      pass

    #-----------------------------------------------------------------------------
    def close_distfile( self ):
      pass

    #-----------------------------------------------------------------------------
    def copy_distfile( self ):
      pass

    #-----------------------------------------------------------------------------
    def remove_distfile( self ):
      pass

    #-----------------------------------------------------------------------------
    def finalize( self ):
      pass

  dist = dist_dummy2('asdasd')
  dist.open()

  with raises(ValueError):
    dist.makedirs('stuff')


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_source_dummy():

  with raises( ValueError ):
    dist_source_dummy( pkg_info = None )

  pkg_info = PkgInfo(
    project = dict(
      name = 'my-package',
      version = '1.0' ) )

  dist_source_dummy( pkg_info = pkg_info )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_targz():

  with tempfile.TemporaryDirectory() as tmpdir:
    with dist_targz(
      outname = 'asd.tgz',
      outdir = tmpdir ) as dist:

      assert not dist.exists('stuff')

      dist.write('stuff', 'stuff content')

      assert dist.exists('stuff')

      dist.assert_recordable()

      dist.finalize()
      dist.finalized = True

      with raises( ValueError ):
        dist.assert_recordable()

    with raises( ValueError ):
      dist.assert_recordable()

    # should be able to call any number of times
    dist.close_distfile()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_zip():

  with tempfile.TemporaryDirectory() as tmpdir:
    with dist_zip(
      outname = 'asd.zip',
      outdir = tmpdir ) as dist:

      assert not dist.exists('stuff')

      dist.write('stuff', 'stuff content')

      assert dist.exists('stuff')

      dist.assert_recordable()

    with raises( ValueError ):
      dist.assert_recordable()

    # should be able to call any number of times
    dist.close_distfile()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_source():

  with raises( ValueError ):
    dist_source_targz( pkg_info = None )

  with tempfile.TemporaryDirectory() as tmpdir:

    pkg_dir = osp.join( tmpdir, 'src', 'my_package' )
    out_dir = osp.join( tmpdir, 'build' )
    mod_file = osp.join( pkg_dir, 'module.py' )

    os.makedirs( pkg_dir )

    with open( mod_file, 'w' ) as fp:
      fp.write("print('hello')")


    pkg_info = PkgInfo(
      project = dict(
        name = 'my-package',
        version = '1.0' ) )

    with dist_source_targz(
      pkg_info = pkg_info,
      outdir = out_dir ) as sdist:

      with raises( ValueError ):
        # already open
        sdist.open()

      sdist.copytree(
        src = osp.join( tmpdir, 'src' ),
        dst = osp.join( sdist.base_path, 'src' ),
        ignore = shutil.ignore_patterns('nothing') )

      with raises( ValueError ):
        # duplicate
        sdist.copytree(
          src = osp.join( tmpdir, 'src' ),
          dst = osp.join( sdist.base_path, 'src' ) )

      sdist.copyfile(
        src = mod_file,
        dst = osp.join( sdist.base_path, 'src', 'mod.py' ) )

      with raises( ValueError ):
        # duplicate
        sdist.copyfile(
          src = mod_file,
          dst = osp.join( sdist.base_path, 'src', 'mod.py' ) )

      with raises( ValueError ):
        # doesn't exist
        sdist.copyfile(
          src = 'asd',
          dst = 'xyz' )

      with raises( ValueError ):
        # doesn't exist
        sdist.copytree(
          src = 'asd',
          dst = 'xyz' )

    # already closed
    sdist.close()

    with raises( ValueError ):
      sdist.copytree(
        src = osp.join( tmpdir, 'src' ),
        dst = osp.join( sdist.base_path, 'src2' ) )

    assert sdist.outname == 'my_package-1.0.tar.gz'
    assert osp.relpath( sdist.outpath, tmpdir ) == 'build/my_package-1.0.tar.gz'
    assert osp.exists(sdist.outpath)

    # overwrite existing file
    sdist = dist_source_targz(
      pkg_info = pkg_info,
      outdir = out_dir )

    sdist.open()

    with sdist:

      sdist.copytree(
        src = osp.join( tmpdir, 'src' ),
        dst = osp.join( sdist.base_path, 'src' ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_dist_binary_wheel():



  with tempfile.TemporaryDirectory() as tmpdir:
    pkg_dir = osp.join( tmpdir, 'src', 'my_package' )
    out_dir = osp.join( tmpdir, 'build' )

    os.makedirs( pkg_dir )

    with open( osp.join( pkg_dir, 'module.py' ), 'w' ) as fp:
      fp.write("print('hello')")


    license_file = osp.join(tmpdir, 'license.rst')

    with open(license_file, 'w') as fp:
      fp.write("my license")


    pkg_info = PkgInfo(
      root = tmpdir,
      project = dict(
        name = 'my-package',
        version = '1.0',
        license = { 'file' : 'license.rst' } ) )

    with raises( ValueError ):
      pkg_info_dynamic = PkgInfo(
        project = dict(
          name = 'my-package',
          version = '1.0',
          dynamic = ['dependencies'] ) )

    with raises( ValueError ):
      dist_binary_wheel( pkg_info = None )

    # with raises( ValueError ):
    #   dist_binary_wheel( pkg_info = pkg_info_dynamic )


    dist_binary_wheel( pkg_info = pkg_info )


    with dist_binary_wheel(
      pkg_info = pkg_info,
      outdir = out_dir,
      compat = [ ( 'py3', 'none', 'any' ), ],
      gen_name = 'custom' ) as bdist:

      for k in bdist.named_dirs.keys():

        bdist.copytree(
          src = pkg_dir,
          dst = bdist.named_dirs[k] + '/my_package' )

    assert bdist.top_level == ['my_package']
    assert bdist.finalize()
    assert bdist.outname == 'my_package-1.0-py3-none-any.whl'
    assert osp.relpath( bdist.outpath, tmpdir ) == 'build/my_package-1.0-py3-none-any.whl'
    assert osp.exists(bdist.outpath)

    # overwrite existing file
    with dist_binary_wheel(
      pkg_info = pkg_info,
      outdir = out_dir,
      compat = [ ( 'py3', 'none', 'any' ), ],
      gen_name = 'custom' ) as bdist:

      bdist.copytree(
        src = pkg_dir,
        dst = 'my_package' )
